#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import os
import paramiko
import sys
import traceback
import yaml
import ConfigParser

from create_user_models import Account, Account_External_Ids, Account_Ssh_Keys
from create_user_models import Account_Group_Members, Group
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import sessionmaker


class User():
    """Basic class to describe a Gerrit user instance."""

    def __init__(self, engine, user, credentials, mode):
        self.engine = engine
        self.user = user
        self.cmd = ''
        self.mode = mode

        self.config = {
            'description': user['description'],
            'email': user['email'],
            'full-name': user['full-name'],
            'group': user['group'],
            'http-password': user['http-password'],
            'login': user['login'],
            'owner': user['owner'],
            'ssh-key': user['ssh-key'],
            'status': user['status'],
        }

        self.host = credentials['host']
        self.port = int(credentials['port'])
        self.user = credentials['user']

        self.login = user['login']

        session = sessionmaker(bind=self.engine)
        self.session = session()

    def _exec(self):
        """User's internal method to execute a command, by using Gerrit's API.
        """

        gerrit = paramiko.SSHClient()
        gerrit.load_system_host_keys()
        gerrit.connect(
            self.host,
            self.port,
            self.user
        )

        try:
            stdin, stdout, stderr = gerrit.exec_command(self.cmd)

            error = stderr.read()

            if not error:
                return stdout.read()
            else:
                logging.error("Can't recieve data from gerrit")
                logging.error(self.cmd)
                logging.error(error)
        except Exception as e:
            logging.error("Can't recieve data from gerrit")
            logging.error('Error: {}'.format(e))
            logging.error(traceback.print_exc(e))

    def check(self):
        """User's method to check does the user exist in Gerrit"""

        query = "username:{}".format(self.login)

        if self.session.query(
                Account_External_Ids
            ).filter(
                    Account_External_Ids.external_id.like(
                        query
                    )
                ).count() == 0:
            return False

        return True

    def operate_account(self):
        """User's method to operate with a User's instance. If self.cmd
            isn't defiend, a new user will be created.
        """
        if not self.cmd:
            self.cmd = '{} {} {} \"{}\"'.format(
                'gerrit create-account',
                self.config['login'],
                '--group',
                self.config['group'][0]
            )

        try:
            if self.mode != 'noop':
                self._exec()
            return True
        except Exception as e:
            logging.error('Error: {}'.format(e))
            logging.error(traceback.print_exc(e))
            return False

    def operate_full_name(self):
        query = "username:{}".format(self.login)
        if self.mode != 'noop':
            login = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.external_id.like(query)
            ).first()

            full_name = self.session.query(
                Account
            ).filter(
                Account.account_id == login.account_id
            )[0].full_name

            if full_name != self.config['full-name']:
                self.cmd = "{} --{} \"{}\" {}".format(
                    "gerrit set-account",
                    "full-name",
                    self.config['full-name'],
                    self.login
                )
                try:
                    self.operate_account()
                    logging.info(
                        "{}: full name is changed to {}".format(
                            self.login,
                            self.config['full-name']
                        )
                    )
                except Exception as e:
                    logging.error('Error: {}'.format(e))
                    logging.error(traceback.print_exc(e))
                    return False
        else:
            logging.info(
                "{}: full name is changed to {}".format(
                    self.login,
                    self.config['full-name']
                )
            )

        return True

    def operate_group(self):
        for i in self.config['group']:
            if i == "Non-Interactive Users":
                if self.login not in get_gerrit_users(self.engine):
                    self.cmd = "{} \"{}\" --{} {}".format(
                        "gerrit set-members",
                        "Non-Interactive Users",
                        "add",
                        self.login
                    )
                    try:
                        self.operate_account()
                        logging.info(
                            "{}: user has been added to {} group".format(
                                self.login,
                                self.config['group'][0]
                            )
                        )
                    except Exception as e:
                        logging.error('Error: {}'.format(e))
                        logging.error(traceback.print_exc(e))
                        return False
            else:
                if get_gerrit_users(self.engine, i) == 'Absent group':
                    self.cmd = "{} {}".format(
                        "gerrit create-group",
                        i
                    )
                    try:
                        self.operate_account()
                        logging.info(
                            "Group {} has been created".format(
                                i
                            )
                        )
                    except Exception as e:
                        logging.error('Error: {}'.format(e))
                        logging.error(traceback.print_exc(e))
                        return False
                if get_gerrit_users(self.engine, i):
                    self.cmd = "{} \"{}\" --{} {}".format(
                        "gerrit set-members",
                        i,
                        "add",
                        self.login
                    )
                    try:
                        self.operate_account()
                        logging.info(
                            "{}: user has been added to {} group".format(
                                self.login,
                                i
                            )
                        )
                    except Exception as e:
                        logging.error('Error: {}'.format(e))
                        logging.error(traceback.print_exc(e))
                        return False
        return True

    def create(self):
        """User's method to create a new user."""

        if self.operate_account():
            logging.info("The user {} created successfully".format(
                self.login
                )
            )

        if self.config['ssh-key'] and len(self.config['ssh-key']) >= 1:
            for i in self.config['ssh-key']:
                self.cmd = "{} {} \"{}\" {}".format(
                    "gerrit set-account",
                    "--add-ssh-key",
                    i,
                    self.login
                )
                if self.operate_account():
                    logging.info(
                        "{}: the following public key added: {}".format(
                            self.login,
                            i
                        )
                    )

        if self.config['email'] and len(self.config['email']) >= 1:
            for i in self.config['email']:
                self.cmd = "{} {} '{}' {}".format(
                    "gerrit set-account",
                    "--add-email",
                    i,
                    self.login
                )
                if self.operate_account():
                    logging.info(
                        "{}: the following email added: {}".format(
                            self.login,
                            i
                        )
                    )

        if self.config['group'] and len(self.config['group']) > 0:
            self.operate_group()

    def set(self):
        """
        User's method to change a user's data: email, group, ssh-key,
        status, full-name.
        """
        if self.config['ssh-key'] and len(self.config['ssh-key']) >= 1:

            query = "username:{}".format(self.login)

            login = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.external_id.like(query)).first()

            q_keys = self.session.query(
                Account_Ssh_Keys
            ).filter(
                Account_Ssh_Keys.account_id == login.account_id
            )

            keys = [i.ssh_public_key for i in q_keys]

            for i in self.config['ssh-key']:
                if i not in keys:
                    self.cmd = "{} {} \"{}\" {}".format(
                        "gerrit set-account",
                        "--add-ssh-key",
                        i,
                        self.login
                    )
                    if self.operate_account():
                        logging.info(
                            "{}: the following public key added: {}".format(
                                self.login,
                                i
                            )
                        )

            for i in keys:
                if i not in self.config['ssh-key']:
                    pass
                    self.cmd = "{} {} \"{}\" {}".format(
                        "gerrit set-account",
                        "--delete-ssh-key",
                        i,
                        self.login
                    )
                    if self.operate_account():
                        logging.info(
                            "{}: the following public key deleted: {}".format(
                                self.login,
                                i
                            )
                        )
        else:
            query = "username:{}".format(self.login)

            login = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.external_id.like(query)
            ).first()

            q_keys = self.session.query(
                Account_Ssh_Keys
            ).filter(
                Account_Ssh_Keys.account_id == login.account_id
            )

            keys = [i.ssh_public_key for i in q_keys]

            for i in keys:
                self.cmd = "{} {} \"{}\" {}".format(
                    "gerrit set-account",
                    "--delete-ssh-key",
                    i,
                    self.login
                )
                if self.operate_account():
                    logging.info(
                        "{}: the following public ssh-key deleted: {}".format(
                            self.login,
                            i
                        )
                    )

        if self.config['email'] and len(self.config['email']) >= 1:
            query = "username:{}".format(self.login)

            login = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.external_id.like(query)
            ).first()

            q_email = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.account_id == login.account_id
            )

            emails = [
                i.email_address for i in q_email if i.email_address is not None
            ]

            if len(emails) > 0:
                for i in self.config['email']:
                    if i not in emails:
                        self.cmd = "{} {} \"'{}'\" {}".format(
                            "gerrit set-account",
                            "--add-email",
                            i,
                            self.login
                        )

                        if self.operate_account():
                            logging.info(
                                "{}: the following email added: {}".format(
                                    self.login,
                                    i
                                )
                            )

                for i in emails:
                    if i not in self.config['email']:
                        self.cmd = "{} {} \"{}\" {}".format(
                            "gerrit set-account",
                            "--delete-email",
                            i,
                            self.login
                        )

                        if self.operate_account():
                            logging.info(
                                "{}: the following email deleted: {}".format(
                                    self.login,
                                    i
                                )
                            )

            else:
                for i in self.config['email']:
                    self.cmd = "{} {} \"{}\" {}".format(
                        "gerrit set-account",
                        "--add-email",
                        i,
                        self.login
                    )

                    if self.operate_account():
                        logging.info(
                            "{}: the following email added: {}".format(
                                self.login,
                                i
                            )
                        )

        else:
            query = "username:{}".format(self.login)

            login = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.external_id.like(query)
            ).first()

            q_email = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.account_id == login.account_id
            )

            emails = [
                i.email_address for i in q_email if i.email_address is not None
            ]

            for i in emails:
                self.cmd = "{} {} \"{}\" {}".format(
                    "gerrit set-account",
                    "--delete-email",
                    i,
                    self.login
                )

                if self.operate_account():
                    logging.info(
                        "{}: the following email deleted: {}".format(
                            self.login,
                            i
                        )
                    )

        if self.config['status']:
            query = "username:{}".format(self.login)

            login = self.session.query(
                Account_External_Ids
            ).filter(
                Account_External_Ids.external_id.like(query)
            ).first()

            inactive = self.session.query(
                Account
            ).filter(
                Account.account_id == login.account_id
            )[0].inactive

            if self.config['status'] == 'active' and inactive == 'Y':
                self.cmd = "{} --{} {}".format(
                    "gerrit set-account",
                    "active",
                    self.login
                )
                if self.operate_account():
                    logging.info(
                        "{} user activated".format(
                            self.login
                        )
                    )
            elif self.config['status'] == 'inactive' and inactive == 'N':
                self.cmd = "{} --{} {}".format(
                    "gerrit set-account",
                    "inactive",
                    self.login
                )
                if self.operate_account():
                    logging.info(
                        "{} user deactivated".format(
                            self.login
                        )
                    )

        if self.config['group'] and len(self.config['group']) > 0:
            self.operate_group()

        elif self.config['group']:
            if self.login not in get_gerrit_users(self.engine):
                self.cmd = "{} \"{}\" --{} {}".format(
                    "gerrit set-members",
                    "Non-Interactive Users",
                    "include",
                    self.login
                )
                if self.operate_account():
                    logging.info(
                        "{}: user has been added to {} group".format(
                            self.login,
                            self.config['group']
                        )
                    )

        if self.config['full-name']:
            self.operate_full_name()

    def manage(self):
        """Basic function to manage by a User's class instance"""

        if not self.check():
            logging.info("User {} does not exists".format(self.login))
            self.create()
        else:
            logging.info("User {} exists".format(self.login))
            self.set()

        self.session.close()


def confparser(config_file=None):
    '''Function to parse the config file

    Args:
        config_file: path to the config file (default: current directory).

    Returns:
        Dict with credentials if successful,
        script exits otherwise.
    '''

    if config_file is None:
        config_file = '{}/{}'.format(os.getcwd(), 'create_user.conf')

    credentials = {}

    if os.path.exists(config_file):
        config = ConfigParser.ConfigParser()
        config.read(config_file)

        for target in config.sections():
            credentials[target] = dict(config.items(target))

        return credentials

    else:
        logging.error(
            "Couldn't find config file ain '{}'".format(
                config_file
            )
        )
        sys.exit(1)


def get_data(path=None):
    '''Function to parse users.yaml file

    Args:
        path: The path to directory where users.yaml file stores.

    Returns:
        Dict of data from yaml file if successful,
        script exits otherwise.
    '''

    if path is None:
        path = '{}/{}'.format(os.getcwd(), 'users.yaml')

    try:
        with open(path, 'r') as f:
            data = yaml.safe_load(f)
        f.close()

        return data
    except Exception as e:
        logging.error(
            "Unexpected error was returned when reading config '{}'".format(
                path
            )
        )
        logging.error('Error: {}'.format(e))
        logging.error(traceback.print_exc(e))
        sys.exit(1)


def get_gerrit_users(engine, ger_group=None):
    """ Function to get a users list from the Gerrit database. As additional
    option, this function can check is there a project in the Gerrit database.

    Args:
        engine: engine to create a session and to connect to Gerrit's database;
        ger_group: name of a group to get.

    Returns:
        Logins if successful,
        Absent group if a group doesn't exists,
        False otherwise.
    """
    if ger_group is None:
        ger_group = "Non-Interactive Users"

    session = sessionmaker(bind=engine)
    session = session()

    logins = []

    try:
        group = session.query(Group).filter(
            Group.name == ger_group)[0]
    except:
        return 'Absent group'

    accounts_id = session.query(Account_Group_Members).filter(
        Account_Group_Members.group_id == group.group_id
    )

    for account in accounts_id:
        login = session.query(Account_External_Ids).filter(
            Account_External_Ids.account_id == account.account_id
        ).filter(
            Account_External_Ids.external_id.like("username:%")
        ).all()[0]
        if login.external_id.replace("username:", "") not in logins:
            logins.append(login.external_id.replace("username:", ""))

    if logins:
        return logins
    else:
        return False


def split_users(cfg_list, db_list):
    """Function to split new users and users which are presented in Gerrit

    Args:
        cfg_list - list of users from configuration file;
        db_list - list of users which are presented in the the Gerrit database.

    Returns:
        List of users if successful, False otherwise.
    """
    new_users = [
        i for i in cfg_list if i['login'] not in db_list
    ]
    if new_users:
        update_users = [
            i for i in cfg_list if i not in new_users
        ]
        return new_users, update_users
    else:
        update_users = [
            i for i in cfg_list
        ]
        return False, update_users


def database_connect(credentials):
    """Function to establish connection to the database.

    Args:
        credentials - dict with parameters to connect to the database:
            drivername
            host
            port
            database
            username
            password

    Returns:
        engine object.
    """
    credentials['drivername'] = 'mysql+mysqldb'

    try:
        engine = create_engine(
            URL(
                drivername=credentials['drivername'],
                username=credentials['username'],
                password=credentials['password'],
                host=credentials['host'],
                port=credentials['port'],
                database=credentials['database']
            ),
            echo=False
        )

    except Exception as e:
        logging.error("A problem with database detected")
        logging.error('Error: {}'.format(e))
        logging.error(traceback.print_exc(e))
        sys.exit(0)

    return engine


def main():
    logging.basicConfig(
        level=logging.INFO,
        format=u'[%(asctime)s] --%(levelname)s-- %(message)s',
    )

    parser = argparse.ArgumentParser(
        description="Script to create new Gerrit's user or update exist users."
    )

    subparsers = parser.add_subparsers(help="The list of main commands.")

    dry_run = subparsers.add_parser(
        'noop',
        help="Dry-run mode. Do nothing, just print information."
    )

    dry_run.add_argument(
        '-c', '--script-config',
        action='store',
        help="Path to the script's config fule."
    )

    dry_run.add_argument(
        '-u', '--user-file',
        action='store',
        help="Path to the config file with users definition."
    )
    dry_run.set_defaults(mode='noop')

    create = subparsers.add_parser(
        'create',
        help="Create new user(s) or update exist users."
    )
    create.add_argument(
        '-c', '--script-config',
        action='store',
        help="Path to the script's config fule."
    )

    create.add_argument(
        '-u', '--user-file',
        action='store',
        help="Path to the config file with users definition."
    )
    create.set_defaults(mode='create')

    args = parser.parse_args()
    mode = args.mode
    config = args.script_config
    users = args.user_file

    credentials = confparser(config_file=config)
    engine = database_connect(credentials['database'])

    db_list = get_gerrit_users(engine)
    cfg_list = get_data(path=users)

    create_user, update_user = split_users(cfg_list, db_list)

    if create_user:
        logging.info("--- Create new users --- ")
        for user in create_user:
            user = User(engine, user, credentials['gerrit'], mode)
            user.manage()
            # The second run should add all user's information.
            # It's not a typo.
            user.manage()

    if update_user:
        logging.info("--- Update of existing users --- ")
        for user in update_user:
            user = User(engine, user, credentials['gerrit'], mode)
            user.manage()


if __name__ == '__main__':
    main()
